"""
Constants
=========

Those constants are generic and can be used across the different datastore
types.
"""

SORT_ASCENDING = 2
SORT_DESCENDING = 1

ERROR_CODE_VALIDATION = 'validation_errors'
ERROR_CODE_DUPLICATE_KEY = 'duplicate_key'



READ = 0
CREATE = 1
UPDATE = 2
DELETE = 3
FIND_UNIQUE = 4
FIND_INDEX = 5
FIND_ONE_PRIMARY = 6


class Base():
    def __init__(self, value):
        self.value = value

    def validate(self, field):
        self.value = field.validate(self.value)
        return self.value


class Equal(Base):
    pass


class GreaterThan(Base):
    pass


class SmallerThan(Base):
    pass


class Contains(Base):
    pass


class Exclude(Base):
    pass


class MultipleInput(Base):
    def __init__(self, *value):
        self.value = value

    def validate(self, field):
        self.value = [field.validate(value) for value in self.value]
        return self.value


class In(MultipleInput):
    pass


class Between(MultipleInput):
    pass



class FieldException(Exception):
    pass

FIELD_ALREADY_USED = FieldException(
    'This is already used.')
FIELD_WRONG_FORMAT = FieldException(
    'This field is not following the right format.')
FIELD_REQUIRED = FieldException(
    'This field is required.')
RESPONSE_FORMAT_ATTRIBUTES_UNAVAILABLE = FieldException(
    'The response format cannot get attributes on a non dictionary message.')



from unittest import mock
import pytest

from sukimu import exceptions
from sukimu import fields


def test_field_creation():
    f = fields.Field()
    assert f
    assert f.basetype == str
    assert f.required is False
    assert len(f.validators) == 0


def test_required_field_creation():
    f = fields.Field(required=True)
    assert f
    assert f.basetype == str
    assert f.required is True
    assert len(f.validators) == 0


def test_field_validators(monkeypatch):
    validator = mock.MagicMock()
    f = fields.Field(validator, basetype=str)
    assert len(f.validators) == 1
    assert f.validators[0] == validator


def test_validation():
    f = fields.Field(required=True)
    wrong_values = (list(), set(), dict())
    assert f.validate('Hello') == 'Hello'
    for wrong_value in wrong_values:
        with pytest.raises(exceptions.FieldException):
            f.validate(wrong_value)


def test_validation_on_required_field():
    f = fields.Field(required=True)
    values = (None, '')
    for value in values:
        with pytest.raises(exceptions.FieldException):
            f.validate(value)


def test_validation_on_wrong_field_type():
    f = fields.Field(basetype=int)
    with pytest.raises(exceptions.FieldException):
        f.validate('value')


def test_validation_with_validators():
    def validator(value):
        return value.replace('hello', 'hi')

    def world_validator(value):
        if 'world' not in value:
            raise exceptions.FieldException(
                'The word "world" has not been found')
        return value.replace('world', 'earth')

    f = fields.Field(validator, world_validator)
    assert f.validate('hello world') == 'hi earth'
    with pytest.raises(exceptions.FieldException):
        f.validate('hello monde')



"""
Field
=====

The field handle the base validation of the values being passed. It checks
first if the value is required, if the type is correct then it invokes any
additional validators.
"""

from sukimu import exceptions


class Field():
    def __init__(
            self, *validators, basetype=str, required=False):
        """Field

        All fields should have a simple basetype (number, str, map) that
        matches a database format.

        Arg:
            validators (set): All the additional validators you will want to
                run against the value.
            basetype (type): A type (str, dict, set) that explains what this
                field is about. The validation will verify if this value is
                working as expected.
            required (boolean): If this field is required. If absent, it will
                throw an exception.
        """

        self.basetype = basetype
        self.required = required
        self.validators = validators or []

    def validate(self, value):
        """Validator

        Perform controls on the value provided.

        Arg:
            value (str): The value to run the validators against.

        Exception:
            FieldException: If an error has occurred with the field value, an
                exception is immediately raised (either by this method, or by
                one of the validator).

        Return:
            The value.
        """

        if not value:
            if self.required:
                raise exceptions.FIELD_REQUIRED
            return value

        if not isinstance(value, self.basetype):
            raise exceptions.FIELD_WRONG_FORMAT

        for validator in self.validators:
            value = validator(value)

        return value



from sukimu import utils


def test_key_dict():
    dictionary = {'key1': 'value1', 'key2': 'value2'}
    resp = utils.key_dict(dictionary, True)
    assert resp.get('key1') is True


def test_dict_from_strings():
    array = ['user.stats.daily', 'user.username']
    response = utils.dict_from_strings(array)
    assert len(response) == 1
    assert len(response.get('user')) == 2
    assert 'username' in response.get('user')
    assert 'stats.daily' in response.get('user')



import time
import pytest
import uuid
from random import random
from random import shuffle

from oto import response
from oto import status

from sukimu import consts
from sukimu import exceptions
from sukimu.dynamodb import IndexDynamo
from sukimu.dynamodb import IndexDynamo
from sukimu.dynamodb import TableDynamo
from sukimu.fields import Field
from sukimu.operations import Between
from sukimu.operations import Equal
from sukimu.operations import In
from sukimu.schema import Schema, Index
from tests.fixtures import dynamodb


@pytest.fixture
def table_name():
    return str(uuid.uuid1())[:8]


@pytest.fixture
def user_schema():
    schema = Schema(
        TableDynamo(table_name(), dynamodb.connection),

        IndexDynamo(
            Index.PRIMARY, 'id', read_capacity=8, write_capacity=2),
        IndexDynamo(
            Index.GLOBAL, 'username', read_capacity=8, write_capacity=2,
            name='username-index'),

        id=Field(),
        username=Field(),
        password=Field(),
        map_field=Field(basetype=dict),
        random_field=Field(basetype=int))
    schema.table.create_table()
    return schema


@pytest.fixture
def thread_schema():
    schema = Schema(
        TableDynamo(table_name(), dynamodb.connection),
        IndexDynamo(Index.PRIMARY, 'forum_name', 'thread_title',
                    read_capacity=8, write_capacity=2),
        IndexDynamo(Index.LOCAL, 'forum_name', 'thread_author',
                    name='local-index'),
        IndexDynamo(Index.GLOBAL, 'thread_title', 'thread_author',
                    name='global-index', read_capacity=8, write_capacity=2),
        forum_name=Field(),
        thread_title=Field(),
        thread_author=Field(),
        thread_content=Field())
    schema.table.create_table()
    return schema


@pytest.fixture
def stats_schema():
    schema = Schema(
        TableDynamo(table_name(), dynamodb.connection),
        IndexDynamo(Index.PRIMARY, 'user_id', 'day_id',
                    read_capacity=8, write_capacity=2),
        user_id=Field(basetype=int),
        day_id=Field(basetype=int),
        metrics=Field(basetype=int))
    schema.table.create_table()
    return schema


@pytest.fixture
def stats_reverse_schema():
    schema = Schema(
        TableDynamo(table_name(), dynamodb.connection),
        IndexDynamo(Index.PRIMARY, 'user_id', 'day_id',
                    read_capacity=8, write_capacity=2),
        IndexDynamo(Index.GLOBAL, 'day_id', 'user_id',
                    name='day-id-user-id', read_capacity=8, write_capacity=2),
        user_id=Field(basetype=int),
        day_id=Field(basetype=int))
    schema.table.create_table()
    return schema


def test_can_create_fixtures(user_schema, thread_schema):
    pass


def test_create_an_entry_with_wrong_field(user_schema):
    resp = user_schema.create(id='30', username='michael', random_field='test')
    assert not resp
    assert isinstance(
        resp.errors.get('message').get('random_field'),
        exceptions.FieldException)

    resp = user_schema.fetch_one(id=Equal('30'))
    assert not resp
    assert resp.status is status.NOT_FOUND


def test_extension(user_schema):
    new_schema = user_schema.extends(
        new_field=Field())
    assert isinstance(new_schema, Schema)
    assert new_schema.table.name == user_schema.table.name
    assert len(new_schema.indexes) == len(user_schema.indexes)
    assert len(new_schema.fields) - 1 == len(user_schema.fields)
    assert not new_schema.table.schema == user_schema


def test_create_an_entry_for_user(user_schema):
    resp = user_schema.create(id='30', username='michael')
    assert resp


def test_update_an_entry_for_user(user_schema):
    resp = user_schema.create(id='30', username='michael')
    assert resp

    resp = user_schema.update(dict(id='30'), username='joe')
    assert resp


def test_delete_an_entry_for_user(user_schema):
    user_schema.create(id='30', username='michael')
    user_schema.create(id='40', username='michael2')

    resp = user_schema.delete(id=Equal('30'))
    assert resp

    resp = user_schema.fetch_one(id=Equal(30))
    assert not resp


def test_update_an_entry_on_existing_key(user_schema):
    user_schema.create(id='.40', username='michael')
    user_schema.create(id='30', username='joe')

    resp = user_schema.update(dict(id='30'), username='michael')
    assert not resp
    assert isinstance(
        resp.errors.get('message').get('username'), exceptions.FieldException)

    resp = user_schema.fetch_one(id=Equal('30'))
    assert resp.message.get('username') == 'joe'


def test_create_an_entry_on_existing_user_id(user_schema):
    resp = user_schema.create(id='30', username='michael')
    assert resp

    resp = user_schema.create(id='30', username='otherusername')
    assert not resp
    assert not resp.errors.get('message').get('username')
    assert resp.errors.get('code') is consts.ERROR_CODE_DUPLICATE_KEY
    assert isinstance(
        resp.errors.get('message').get('id'), exceptions.FieldException)


def test_create_an_entry_with_map_data(user_schema):
    resp = user_schema.create(
        id='190',
        username='emichael90',
        map_field=dict(
            key1='value',
            key2=dict(
                key1='value')))
    assert resp

    resp = user_schema.fetch_one(id=Equal('190'))
    assert resp
    assert isinstance(resp.message.get('map_field'), dict)


def test_create_an_entry_on_existing_user_username(user_schema):
    resp = user_schema.create(id='30', username='michael')
    assert resp

    resp = user_schema.create(id='20', username='michael')
    assert not resp
    assert not resp.errors.get('message').get('id')
    assert resp.errors.get('code') is consts.ERROR_CODE_DUPLICATE_KEY
    assert isinstance(
        resp.errors.get('message').get('username'), exceptions.FieldException)


def test_create_an_entry_on_existing_user_username_and_id(user_schema):
    resp = user_schema.create(id='30', username='michael')
    assert resp

    resp = user_schema.create(id='20', username='michael')
    assert not resp
    assert not resp.errors.get('message').get('id')
    assert resp.errors.get('code') is consts.ERROR_CODE_DUPLICATE_KEY
    assert isinstance(
        resp.errors.get('message').get('username'), exceptions.FieldException)


def test_thread_creation(thread_schema):
    resp = thread_schema.create(
        forum_name='News', thread_title='title', thread_author='user',
        thread_content='content')
    assert resp

    resp = thread_schema.fetch_one(
        forum_name=Equal('News'), thread_title=Equal('title'))
    assert resp
    assert resp.message.get('thread_author') == 'user'

    resp = thread_schema.fetch_one(
        forum_name=Equal('News'), thread_author=Equal('user'))
    assert resp
    assert resp.message.get('thread_title') == 'title'

    resp = thread_schema.fetch_one(
        thread_title=Equal('title'), thread_author=Equal('user'))
    assert resp
    assert resp.message.get('forum_name') == 'News'

    resp = thread_schema.create(
        forum_name='Updates', thread_title='Title2', thread_author='user',
        thread_content='content')
    assert resp

    resp = thread_schema.create(
        forum_name='Updates', thread_title='Title3', thread_author='user2',
        thread_content='content')
    assert resp

    resp = thread_schema.create(
        forum_name='Others', thread_title='Title', thread_author='user4',
        thread_content='foobar')
    assert resp


def test_thread_creation_on_duplicate_indexes(thread_schema):
    # Indexes:
    # - Forum Name and Thread Title
    # - Forum Name - Author
    # - Forum Title - Author
    resp = thread_schema.create(
        forum_name='News', thread_title='title', thread_author='user',
        thread_content='content')
    assert resp

    resp = thread_schema.create(
        forum_name='News', thread_title='title', thread_author='user2',
        thread_content='content')
    assert not resp
    assert resp.errors.get('message').get('forum_name')
    assert resp.errors.get('message').get('thread_title')

    resp = thread_schema.create(
        forum_name='News', thread_title='title2', thread_author='user',
        thread_content='content')
    assert not resp
    assert resp.errors.get('message').get('thread_author')
    assert resp.errors.get('message').get('forum_name')

    resp = thread_schema.create(
        forum_name='Other', thread_title='title', thread_author='user',
        thread_content='content')
    assert not resp
    assert resp.errors.get('message').get('thread_title')
    assert resp.errors.get('message').get('thread_author')


def test_create_dynamo_schema(table_name):
    table = TableDynamo(table_name, dynamodb.connection)
    primary_index = IndexDynamo(Index.PRIMARY, 'id')
    global_index = IndexDynamo(Index.GLOBAL, 'foo', 'bar')
    tb = Schema(table, primary_index, global_index)

    assert tb.table == table
    assert tb.indexes[0] == primary_index
    assert tb.indexes[1] == global_index


def test_fetch_on_index(thread_schema):
    resp = thread_schema.create(
        forum_name='News', thread_title='title', thread_author='user',
        thread_content='content')
    assert resp

    resp = thread_schema.fetch(
        forum_name=Equal('News'), thread_title=Equal('title'))
    assert resp
    assert resp.message[0].get('thread_author') == 'user'

    resp = thread_schema.fetch(
        thread_title=Equal('title'), thread_author=Equal('user'))
    assert resp
    assert resp.message[0].get('forum_name') == 'News'


def test_fetch_many(user_schema):
    user_schema.create(id='30', username='michael1')
    user_schema.create(id='40', username='michael2')
    resp = user_schema.fetch(username=In('michael1', 'michael2'))
    assert resp
    assert len(resp.message) == 2


def test_between_request(stats_schema):
    stats_schema.create(user_id=301, day_id=35, metrics=937)

    for day in range(50):
        metrics = int(random() * 400)
        resp = stats_schema.create(user_id=300, day_id=day, metrics=metrics)

    resp = stats_schema.fetch(user_id=Equal(300), day_id=Between(30, 40))
    assert len(resp.message) == 11  # 40 is included


def test_sorting(stats_schema):
    days = list(range(50))
    shuffle(days)
    for day in days:
        metrics = int(random() * 400)
        resp = stats_schema.create(user_id=300, day_id=day, metrics=metrics)

    resp = stats_schema.fetch(user_id=Equal(300), sort=consts.SORT_DESCENDING)
    start = 49
    for i in range(50):
        assert resp.message[i].get('day_id') == start
        start = start - 1

    resp = stats_schema.fetch(user_id=Equal(300), sort=consts.SORT_ASCENDING)
    for i in range(50):
        assert resp.message[i].get('day_id') == i


def test_reverse_schema(stats_reverse_schema):
    days = list(range(50))
    shuffle(days)
    total_reverse = 0
    for day in days:
        metrics = int(random() * 400)
        stats_reverse_schema.create(
            user_id=300, day_id=day, metrics=metrics)

        if not day % 2:
            total_reverse += 1
            stats_reverse_schema.create(
                user_id=200, day_id=day, metrics=metrics)

    resp = stats_reverse_schema.fetch(user_id=Equal(300))
    assert len(resp.message) == 50

    resp = stats_reverse_schema.fetch(user_id=Equal(200))
    assert len(resp.message) == total_reverse

    resp = stats_reverse_schema.fetch(day_id=Equal(18))
    assert len(resp.message) == 2


def test_dynamo_table_creation(table_name):
    tb = Schema(
        TableDynamo(table_name, dynamodb.connection),
        IndexDynamo(
            Index.PRIMARY, 'id', read_capacity=8, write_capacity=4),
        id=Field())

    tb.table.create_table()
    assert table_name in dynamodb.connection.list_tables().get('TableNames')


def test_dynamo_table_creation_collision(table_name):
    tb = Schema(
        TableDynamo(table_name, dynamodb.connection),
        IndexDynamo(
            Index.PRIMARY, 'id', read_capacity=8, write_capacity=2),
        id=Field())
    tb.table.create_table()

    with pytest.raises(Exception):
        tb.table.create_table()


def test_create_empty_table(table_name):
    """Test the creation of an empty table.
    """
    tb = Schema(TableDynamo(table_name, dynamodb.connection))

    with pytest.raises(Exception):
        tb.table.create_table()


def test_create_table_without_index(table_name):
    tb = Schema(TableDynamo(table_name, dynamodb.connection), id=Field())
    with pytest.raises(Exception):
        tb.table.create_table()


def test_create_table_without_fields(table_name):
    tb = Schema(
        TableDynamo(table_name, dynamodb.connection),
        IndexDynamo(Index.PRIMARY, 'id', read_capacity=1, write_capacity=1))

    with pytest.raises(Exception):
        tb.table.create_table()


def test_extension_usage(user_schema):
    @user_schema.extension('stats')
    def stats(item, fields):
        return {'days': 10, 'fields': fields}

    @user_schema.extension('history')
    def history(item, fields):
        return {'length': 20}

    response = user_schema.create(id='testextension', username='michael')
    assert response

    response = user_schema.fetch_one(
        username=Equal('michael'), fields=['stats.foobar', 'stats.tests.bar'])
    assert response.message.get('stats').get('days') == 10
    assert 'foobar' in response.message.get('stats').get('fields')
    assert 'tests.bar' in response.message.get('stats').get('fields')

    response = user_schema.fetch_one(
        username=Equal('michael'),
        fields=['history', 'stats.foobar', 'stats.tests.bar'])
    assert response.message.get('stats').get('days') == 10
    assert 'foobar' in response.message.get('stats').get('fields')
    assert 'tests.bar' in response.message.get('stats').get('fields')
    assert response.message.get('history').get('length') == 20



def key_dict(dictionary, default=None):
    """Create a dictionary with default keys.

    This methods duplicate the keys of the entered dictionary and recreate a
    new dictionary with the same keys and for default the default value
    provided.

    Example:

        keys = {'user': Field(), 'password': Field()}
        key_dict(keys, 'hello')
        ↳ {'user': 'hello', 'password': 'hello'}

    Args:
        dictionary (dict): The dictionary to treat.
        default: The default value for each of the keys in the new dictionary.
    """

    return {key: default for key in dictionary.keys()}


def key_exclude(dictionary, keys):
    """Exclude all the keys of a dictionary.

    This method removes all the keys from a dictionary that have been passed
    as a second parameter.

    Args:
        dictionary (dict): The used dictionary.
        keys (list): The keys of the dictionary.

    Return:
        Dictionary: The dictionary without the keys.
    """

    return {key: val for key, val in dictionary.items() if key not in keys}


def dict_from_strings(array, separator='.'):
    """Recreate a dictionary from strings that are in an array.

    If you have strings such as "user.id", "user.name", those ideally can be
    turned into a dictionary where user is the principal key and name, id two
    values of this dictionary.

    Args:
        array (array): The array to turn into a dict.
    Return:
        dict: The dictionary.
    """

    dictionary = {}

    for string in array:
        parts = string.split(separator)
        key = parts[0]
        last_parts = separator.join(parts[1:])

        if last_parts:
            dictionary.setdefault(key, []).append(last_parts)
        else:
            dictionary.update({key: []})

    return dictionary


def dict_merge(*dicts):
    """Merge all dictionnary into one.

    In the event of a collision, the last dictionary is the one that will
    override the initial value.

    Args:
        dicts (list): List of all the dictionary to merge.
    Return:
        dict: The dictionary that contains all the other ones.
    """

    dictionaries = list()
    for d in dicts:
        dictionaries = dictionaries + list(d.items())

    return dict(dictionaries)


def list_merge(*lists):
    """Merge all the lists into one.

    Args:
        lists (list): List of all the lists to merge.
    Return:
        list: the result of the merged lists.
    """

    final_list = []
    for l in lists:
        final_list = final_list + l
    return final_list



from setuptools import find_packages
from setuptools import setup


setup(
    name='Sukimu',
    version='1.0.1',
    url='https://github.com/xethorn/sukimu',
    author='Michael Ortali',
    author_email='github@xethorn.net',
    description=(
        'Standardized way to perform CRUD operations with Field validation'),
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: Alpha',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],)



"""
Schemas
=======

Schemas defines the structure of the fields of your table. The schema handles
fields validations, and ensure index unicity. There are 5 main concepts around
the schemas:

**Indexes**

An index defines which key (or set of keys) should be unique within your table.
The schema will perform checks on those indexes whenever a row is being created
or updated.

Some examples:

1. If you have a user table, and need usernames and emails to be unique, you
   will have then 2 indexes.
2. If you have a session token table with a user id and a token number, you can
   have one index composed of two keys: user id (hash) and token number (range)

**Validators**

Validators are health checks on the provided data. For example: if you have a
field `age`, the field is most likely going to have a defined range (minimum
and maximum). If a value provided is not valid, the field validator throws an
exception, caught by the schema, and returned as part of the response (so if
more than one field is invalid, the user can be informed.)

**Extensions**

Extensions are only available for `fetch` and `fetch_one` method. They are
populating more fields on the returned object.

For instance: if you have a user table, and a profile table, you probably want
the user to be able to get the profile as part of the same response. So
`profile` can be an extension of user.

**Generated fields**

Generated fields are created after field validation. For instance: on a blog
post, you want to capture the number of words, it could be a generated field.
Those fields are saved into the database.

**Operations**

Field operations are used for two things: the first is to validate all the
possible value this operation contains (by using the field itself) and the
second is to write the correct query.

For example: ``fetch_one(dict(username=Equal('michael')))``, will convert the
key into ```username__eq``` in dynamodb.

**Some examples**

Create a basic schema::

    user = models.Schema(table, **fields)

Create a schema that has generated fields:

    user = models.Schema(table, **fields)

    @user.generated('username')
    def add_username_to_hisory(model):
        # do something.
        return model

Create extensions::

    user = models.Schema(collection, **fields)

    @user.extension('messages')
    def profile(obj):
        # get the profile for the user.
        return profile
"""

from collections import namedtuple
from copy import deepcopy
from threading import Thread

from oto import response
from oto import status

from sukimu import consts
from sukimu import exceptions
from sukimu import operations
from sukimu import utils


class Schema():

    def __init__(self, table, *indexes, **fields):
        """Initialize the Schema.

        Args:
            table (string): Name of the table or string.
            options (dict): All the schema options.
        """

        self.table = table
        self.indexes = indexes
        self.fields = fields

        self.fields_dependencies = utils.key_dict(fields, default=[])
        self.extensions = dict()

        # Register the schema with the table.
        self.table.set_schema(self)

    def validate(self, values, operation):
        """Validate the model.

        Args:
            values (dict): The values to validate.
            operation (int): The different operations
                (correspond to schema operations).
        Return:
            Response.
        """

        data = dict()
        errors = dict()
        success = False
        items = set(values.keys())

        if operation is operations.READ and not values:
            return response.Response()

        if operation is operations.CREATE:
            items = set(self.fields.keys())

        for name in items:
            field = self.fields.get(name)
            if not field:
                continue

            try:
                value = values.get(name)

                if isinstance(value, operations.Base):
                    value = value.validate(field)
                    data[name] = value
                    continue

                data[name] = field.validate(value)

            except exceptions.FieldException as e:
                errors[name] = e
                status = False

        if errors:
            return response.create_error_response(
                consts.ERROR_CODE_VALIDATION, errors)

        return response.Response(message=data)

    def ensure_indexes(self, validation_response, current=None):
        """Ensure index unicity.

        One particularity of an index: it should be unique. For example in
        DynamoDb: an index has a hash and a range, combined, they represent the
        key of that specific row – and thus should be unique.

        If an index only has one key, that key on its own should also be
        unique (e.g. user_id).

        Args:
            validation_response (Response): The validation response that
                contains the validated fields.
            current (dict): Operations such as update requires to check against
                the found ancestor and the current row that needs to be
                validated.
        Return:
            Response: The response
        """

        if not validation_response:
            return validation_response

        data = validation_response.message
        errors = {}
        current = current or {}

        for index in self.indexes:
            # Some databases allow to have non unique indexes. In this case,
            # we ignore this index for the check.
            if not index.unique:
                continue

            keys = index.keys

            query = dict()
            for key in keys:
                key_value = data.get(key, current.get(key))
                if not key_value:
                    break
                query.update({key: operations.Equal(key_value)})

            if not query:
                continue

            ancestor = self.fetch_one(**query)
            if ancestor:
                if not current or dict(ancestor.message) != dict(current):
                    errors.update({
                        key: exceptions.FIELD_ALREADY_USED for key in keys})

        if errors:
            return response.create_error_response(
                consts.ERROR_CODE_DUPLICATE_KEY, errors)

        return response.Response()

    def generated(self, **dependencies):
        """Register a generated field.

        Generated fields may have some dependencies. If a specific has been
        updated for instance, the generated field will also need to be updated.
        If the generated field needs to be updated everytime.

        Args:
            dependencies (dict): The dependencies for this specific generated
                field.

        Return:
            Response: the response with the generated value.
        """

        return NotImplemented

    def extension(self, name):
        """Register an extension.

        Args:
            name (string): Name of the extension.
        """

        def wrapper(method):
            self.extensions.update({name: method})
            return method
        return wrapper

    def fetch(self, fields=None, limit=None, sort=None, index=None,
                context=None, **query):
        """Query the table to find all the models that correspond to the query.

        Args:
            fields (list): the list of fields to return on each of the items.
            limit (int): optional limit on how many items need to be fetched.
            sort (int): if the results should be sorted, and if so, in which
                order.
            index (str): name of the index to use.
            context (dict): additional context to provide (used by extensions)
            query (dict): fields to query on.

        Return:
            Response: the data of the request.
        """

        validation_response = self.validate(query, operation=operations.READ)
        if not validation_response:
            return validation_response

        schema_response = self.table.fetch(
            query, sort=sort, limit=limit, index=index)
        if schema_response and fields:
            self.decorate_response(schema_response, fields, context=context)

        return schema_response

    def fetch_one(self, fields=None, context=None, **query):
        """Fetch one specific item.

        Args:
            fields (list): the list of fields to return on the item.
            query (dict): the request fields to search on.
            context (dict): optional context (used by extensions).

        Return:
            Response: the data from the request.
        """

        validation_response = self.validate(query, operation=operations.READ)

        if not validation_response:
            return validation_response

        schema_response = self.table.fetch_one(**query)
        if schema_response and fields:
            self.decorate_response(schema_response, fields, context=context)

        return schema_response

    def decorate_response(self, response, fields, context=None):
        """Decorate a response.

        Args:
            item (dict): The current item.
            fields (dict): The fields that are need to be provided to the main
                item.
            context (dict): Additional context to provide to each extension.

        Return:
            Response: the decorated response.
        """

        if (isinstance(fields, list) or isinstance(fields, tuple) or
                isinstance(fields, set)):
            fields = utils.dict_from_strings(fields)

        data = response.message
        if isinstance(data, list):
            data = [
                self.decorate(dict(item), fields, context) for item in data]
        else:
            data = self.decorate(dict(data), fields, context)
        response.message = data

    def decorate(self, item, fields, context=None):
        """Decorate an item with more fields.

        Decoration means that some fields are going to be added to the initial
        item (using the extension with the same name.) The fields that are
        expected from this extension are also being passed.

        Fields are also cleaning the response object (unless unspecified.) For
        instance if you fetch one `user` with the fields `user.id`, only the id
        will be returned.

        Args:
            item (dict): The current item.
            fields (dict): The fields that are need to be provided to the main
                item.
            context (dict): Additional context to provide to each extension.

        Return:
            Response: the decorated response.
        """

        def activate_extension(field, item, context=None):
            extension = self.extensions.get(field)
            if not extension:
                return

            kwargs = {}
            if 'context' in extension.__code__.co_varnames:
                kwargs.update(context=context)
            item.update({field: extension(item, fields.get(field), **kwargs)})

        table_fields = fields.pop(self.table.name, -1)

        threads = []
        for field in fields:
            thread = Thread(
                target=activate_extension, args=(field, item, context))
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

        if table_fields == -1:
            return item

        keys = list(item.keys())
        for key in keys:
            if key not in table_fields:
                item.pop(key)

            if len(item) == len(table_fields):
                return item
        return item

    def create(self, **data):
        """Create a model from the data passed.
        """

        validation = self.validate(data, operation=operations.CREATE)
        if not validation:
            return validation

        check = self.ensure_indexes(validation)
        if not check:
            return check

        data = self.table.create(validation.message)
        return response.Response(message=data)

    def update(self, source, **data):
        """Update the model from the data passed.
        """

        if not source:
            return response.create_error_response(
                message='The source cannot be empty.')

        data = utils.key_exclude(data, source.keys())
        data = self.validate(data, operation=operations.READ)
        if not data:
            return data

        # Recreate the object - check ancestors.
        current = self.fetch_one(**{
            key: operations.Equal(val) for key, val in source.items()})
        if not current:
            return current

        fields = response.Response(
            message=dict(list(source.items()) + list(data.message.items())))

        ancestors = self.ensure_indexes(fields, current.message)
        if not ancestors:
            return ancestors

        return self.table.update(current, fields.message)

    def delete(self, **source):
        """Delete the model(s) from the data passed.
        """

        item = self.fetch_one(**source)
        if not item:
            return item

        return self.table.delete(item.message)

    def extends(self, **fields):
        """Extending a Schema.

        Extension of a schema allows to add new fields. If you have a table
        with users, some users might require different fields (for instance,
        if the user has a gaming console, you might want to get more details
        about this gaming console.)
        """

        fields = utils.dict_merge(self.fields, fields)
        table = self.table.copy()
        indexes = deepcopy(self.indexes)
        return Schema(table, *indexes, **fields)


class Table():

    def __init__(self, name):
        self.name = name
        self.indexes = {}

    def set_schema(self, schema):
        self.schema = schema
        for index in self.schema.indexes:
            self.add_index(index)

    def add_index(self, index):
        return NotImplemented

    def find_index(self, fields):
        for index in self.indexes.values():
            if len(fields) == 2:
                is_hash = index.hash in (fields[0], fields[1])
                is_range = index.range in (fields[0], fields[1])
                if is_hash and is_range:
                    return index
            if len(fields) == 1 and index.hash == fields[0]:
                return index

    def create(self, data):
        return NotImplemented

    def delete(self, source):
        return NotImplemented

    def update(self, source, data):
        return NotImplemented

    def fetch(self, query, sort=None, limit=None):
        return NotImplemented

    def fetch_one(self, **query):
        return NotImplemented

    def create_table(self):
        return NotImplemented

    def copy(self):
        return NotImplemented


class Index():
    PRIMARY = 1
    LOCAL = 2
    GLOBAL = 3

    def __init__(self, *keys, name=None, unique=True):
        self.name = name
        self.keys = list(keys)
        self.unique = unique



"""Adaptor for DynamoDB.
"""

from boto.dynamodb2 import table
from oto import response
from oto import status

from sukimu import consts
from sukimu import operations
from sukimu import schema


class TableDynamo(schema.Table):

    def __init__(self, name, connection):
        """Create a TableDynamo.

        Args:
            name (string): Name of the table.
            connection (DynamoDBConnection): The dynamodb connection.
        """

        self.name = name
        self.connection = connection
        self.table = table.Table(name, connection=connection)
        self.table.use_boolean()  # This allows to use boolean values.
        self.indexes = {}
        self.hash = None
        self.range = None

    def copy(self):
        """Create a copy of the current object.

        Return:
            TableDynamo: Copy of the current instance.
        """

        return self.__class__(self.name, self.connection)

    def add_index(self, index):
        """Add an index into the tbale.

        Args:
            index (DynamoIndex): The dynamo index.
        """

        if index.type == schema.Index.PRIMARY:
            self.hash = index.hash
            self.range = index.range

        self.indexes.update({index.name: index})

    def create(self, data):
        """Create an item.

        Args:
            data (dict): Data for this specific item (refer to the boto
                dynamodb documentation to know what values are allowed.)
        Return:
            Dictionnary: The data that was added.
        """

        self.table.put_item(data=data)
        return data

    def update(self, item, data):
        """Update an item.

        Args:
            item (DynamoDbItem): The dynamodb item to update.
            data (object): The validated fields.
        Return:
            Response: The response of the update.
        """

        if not item:
            return item

        item = item.message
        for field, value in data.items():
            item[field] = value

        save = item.partial_save()
        return response.Response(
            status=status.OK if save else status.ACCEPTED,
            message=item)

    def delete(self, item):
        """Delete an item.

        Args:
            item (DynamoDbItem): The dynamodb item to update.
        Return:
            Response: the response of the update.
        """

        deleted = item.delete()
        return response.Response(
            status=status.OK if deleted else status.BAD_REQUEST)

    def fetch(self, query, sort=None, limit=None, index=None):
        """Fetch one or more entries.

        Fetching entries is allowed on any field. For better performance, it is
        recommended to use one of the indexes. If no index is used, a scan will
        be performed on the table (which are much slower.)

        Args:
            query (dict): the query.
            sort (int): the sorting type (refer to schema.sort).
            limit (int): the number of items you want to get back from the
                table.
            index (str): the name of the index to used. If defined, looking for
                it.
        Return:
            List: All the fetched items.
        """

        data = dict()
        keys = list(query.keys())

        if index:
            index = self.indexes.get(index)
            assert index, 'The index requested was not found.'
        else:
            index = self.find_index(keys)

        if limit:
            data.update(limit=limit)

        for query_segment in query.items():
            key, value = query_segment

            if isinstance(value, operations.Equal):
                data[key + '__eq'] = value.value

            elif isinstance(value, operations.Between):
                data[key + '__between'] = value.value

            elif isinstance(value, operations.In):
                if index:
                    return self.fetch_many(key, value.value, index=index.name)
                return self.fetch_many(key, value.value, index=index.name)

        if index:
            if index.name:
                data['index'] = index.name

            data['reverse'] = sort is consts.SORT_DESCENDING
            dynamo = list(self.table.query_2(**data))
        else:
            dynamo = list(self.table.scan(**data))

        if not len(dynamo):
            return response.Response(
                status=status.NOT_FOUND,
                message=[])

        return response.Response(
            message=[obj for obj in dynamo])

    def fetch_many(self, key, values, index=None):
        """Fetch more than one item.

        Method used to fetch more than one item based on one key and many
        values.

        Args:
            key (string): Name of the key.
            values (list): All the values to fetch.
            index (str): the name of the index to used. If defined, looking for
                it.
        """

        message = []
        for value in values:
            message.append(
                self.fetch_one(
                    index=index, **{key: operations.Equal(value)}).message)

        data_status = status.OK
        if not message:
            data_status = status.NOT_FOUND

        return response.Response(status=data_status, message=message)

    def fetch_one(self, index=None, **query):
        """Get one item.

        Args:
            query (dict): The query item.
            index (str): the name of the index to used. If defined, looking for
                it.

        Return:
            Response: If the item is found, it is provided in the message,
                if not found, the status is set to NOT_FOUND.
        """

        default_response = response.Response(status=status.NOT_FOUND)
        field_names = list(query.keys())

        required = 1
        is_hash = self.hash in field_names
        is_range = self.range in field_names
        if is_range:
            required = 2

        item = None

        if len(query) == required and is_hash:
            data = dict()
            data[self.hash] = query.get(self.hash).value

            if is_range:
                data[self.range] = query.get(self.range).value

            try:
                item = self.table.get_item(**data)
            except:
                return default_response

        if not item:
            data = self.fetch(query, index=index, limit=1).message
            if data and len(data) == 1:
                item = data[0]

        if item:
            return response.Response(message=item)

        return default_response

    def create_table(self):
        """Create a dynamodb table.

        The dynamodb table only needs to know about the indexes and the type
        of those indexes.
        """

        local_secondary_index = []
        global_secondary_index = []

        attributes = []
        attribute_keys = set()
        indexes = {}
        indexes[schema.Index.PRIMARY] = None

        table_provision = None

        for index in self.schema.indexes:
            hash_field = self.schema.fields.get(index.hash)
            range_field = self.schema.fields.get(index.range)

            provision = dict(
                ReadCapacityUnits=index.read_capacity,
                WriteCapacityUnits=index.write_capacity)

            if index.hash not in attribute_keys:
                attribute_keys.add(index.hash)
                attributes.append(dict(
                    AttributeName=index.hash,
                    AttributeType=DynamoType.get(hash_field.basetype)))

            key_schema = [dict(
                AttributeName=index.hash,
                KeyType='HASH')]

            if range_field:
                if index.range not in attribute_keys:
                    attribute_keys.add(index.range)
                    attributes.append(dict(
                        AttributeName=index.range,
                        AttributeType=DynamoType.get(range_field.basetype)))

                key_schema.append(dict(
                    AttributeName=index.range,
                    KeyType='RANGE'))

            if index.type == schema.Index.PRIMARY:
                table_provision = provision
                indexes[index.type] = key_schema
                continue

            response = dict(
                IndexName=index.name,
                KeySchema=key_schema,
                Projection={'ProjectionType': 'ALL'})

            if index.type == schema.Index.GLOBAL:
                response.update(ProvisionedThroughput=provision)

            indexes.setdefault(index.type, []).append(response)

        self.table.connection.create_table(
            attributes, self.name, indexes.get(schema.Index.PRIMARY),
            table_provision,
            local_secondary_indexes=indexes.get(schema.Index.LOCAL, None),
            global_secondary_indexes=indexes.get(schema.Index.GLOBAL, None))


class IndexDynamo(schema.Index):

    def __init__(
            self, type, *keys, name=None, read_capacity=None,
            write_capacity=None, unique=True):
        self.name = name
        self.type = type
        self.hash = keys[0]
        self.range = keys[1] if len(keys) == 2 else None
        self.read_capacity = read_capacity
        self.write_capacity = write_capacity
        self.unique = unique

    @property
    def keys(self):
        keys = [self.hash]
        if self.range:
            keys.append(self.range)
        return keys

DynamoType = {
    str: 'S',
    int: 'N',
}



from unittest import mock
import pytest

from oto import response
from oto import status

from sukimu import fields
from sukimu import operations
from sukimu import schema


def create_schema(fields=None, indexes=None):
    table = schema.Table('TableName')
    indexes = indexes or set()
    fields = fields or dict()
    return schema.Schema(table, *indexes, **fields)


def test_create_schema(monkeypatch):
    table = schema.Table('TableName')
    s = schema.Schema(table)
    assert isinstance(s, schema.Schema)
    assert s.table is table
    assert table.schema == s


def test_create_schema_with_fields():
    table = schema.Table('TableName')
    s = schema.Schema(table, id=fields.Field())
    assert isinstance(s.fields_dependencies.get('id'), list)
    assert len(s.indexes) == 0


def test_create_schema_with_indexes(monkeypatch):
    table = schema.Table('TableName')
    index = schema.Index('id')
    monkeypatch.setattr(table, 'add_index', mock.MagicMock())
    s = schema.Schema(table, index)

    assert len(s.indexes) == 1
    assert table.add_index.called


def test_field_validation_on_create():
    s = create_schema(fields=dict(
        id=fields.Field(),
        username=fields.Field(required=True)))

    # operation is mandatory
    with pytest.raises(Exception):
        s.validate({})

    resp = s.validate({'id': 'test'}, operations.CREATE)
    assert resp.status is status.BAD_REQUEST
    assert resp.errors.get('message').get('username')
    assert not resp.errors.get('message').get('id')

    resp = s.validate({'username': 'test'}, operations.CREATE)
    assert resp.status is status.OK

    resp = s.validate({}, operations.READ)
    assert resp.status is status.OK


def test_field_validation_on_read():
    s = create_schema(fields=dict(
        id=fields.Field(),
        username=fields.Field(required=True)))

    resp = s.validate(
        {'username': 'foo', 'unknownfield': 'value'}, operations.READ)
    assert resp.status is status.OK
    assert not resp.message.get('unknownfield')

    # Fields to validate should be a dictionary of format:
    # <field name, value>
    with pytest.raises(Exception):
        s.validate([], operations.READ)


@pytest.fixture
def full_schema():
    return create_schema(
        indexes=[
            schema.Index('id'),
            schema.Index('id', 'username')],
        fields=dict(
            id=fields.Field(),
            username=fields.Field(required=True)))


def test_ensure_index(monkeypatch, full_schema):
    # If the validation_response is not the rigt object, throws an exception.
    with pytest.raises(Exception):
        full_schema.ensure_indexes(object())

    error_response = response.Response(status=status.BAD_REQUEST)
    assert full_schema.ensure_indexes(error_response) is error_response

    data = dict(id='id-value', username='username-value')
    fetch_one = mock.MagicMock(return_value=error_response)
    success_response = response.Response(data)
    monkeypatch.setattr(full_schema, 'fetch_one', fetch_one)

    resp = full_schema.ensure_indexes(success_response)
    assert resp

    fetch_one.return_value = success_response
    resp = full_schema.ensure_indexes(success_response)
    assert not resp
    assert 'id' in resp.errors.get('message')
    assert 'username' in resp.errors.get('message')


def test_extensions(full_schema):
    @full_schema.extension('stats')
    def stats(item, fields):
        return

    assert full_schema.extensions.get('stats')


def test_decorating_with_extension(full_schema):
    """Test decorating with an extension.
    """

    spy = mock.MagicMock()
    item = {'id': 'foo'}
    context = {'value': 'context_value'}
    fields = {'extension_name': ['foo']}
    extension_name = 'extension_name'

    @full_schema.extension(extension_name)
    def stats(item, fields, context=None):
        spy(item, fields, context)
        return context.get('value')

    response = full_schema.decorate(item, fields=fields, context=context)
    spy.assert_called_with(item, ['foo'], context)
    assert spy.called
    assert response.get(extension_name) == context.get('value')
